@@ -19,6 +19,7 @@ ldns_dnssec_data_chain *
 ldns_dnssec_data_chain_new()
 {
 	ldns_dnssec_data_chain *nc = LDNS_XMALLOC(ldns_dnssec_data_chain, 1);
+        if(!nc) return NULL;
 	nc->rrset = NULL;
 	nc->parent_type = 0;
 	nc->parent = NULL;
@@ -107,6 +108,7 @@ ldns_dnssec_build_data_chain_dnskey(ldns_resolver *res,
 									LDNS_RR_TYPE_DNSKEY,
 									c,
 									qflags);
+			if (my_pkt) {
 			keys = ldns_pkt_rr_list_by_name_and_type(
 					  my_pkt,
 					 key_name,
@@ -120,6 +122,7 @@ ldns_dnssec_build_data_chain_dnskey(ldns_resolver *res,
 													NULL);
 			new_chain->parent->packet_qtype = LDNS_RR_TYPE_DNSKEY;
 			ldns_pkt_free(my_pkt);
+			}
 		} else {
 			new_chain->parent = ldns_dnssec_build_data_chain(res,
 													qflags,
@@ -160,6 +163,7 @@ ldns_dnssec_build_data_chain_other(ldns_resolver *res,
 							LDNS_RR_TYPE_DS,
 							c,
 							qflags);
+	if (my_pkt) {
 	dss = ldns_pkt_rr_list_by_name_and_type(my_pkt,
 									key_name,
 									LDNS_RR_TYPE_DS,
@@ -175,12 +179,14 @@ ldns_dnssec_build_data_chain_other(ldns_resolver *res,
 		ldns_rr_list_deep_free(dss);
 	}
 	ldns_pkt_free(my_pkt);
+	}
 
 	my_pkt = ldns_resolver_query(res,
 							key_name,
 							LDNS_RR_TYPE_DNSKEY,
 							c,
 							qflags);
+	if (my_pkt) {
 	signatures2 = ldns_pkt_rr_list_by_name_and_type(my_pkt,
 										   key_name,
 										   LDNS_RR_TYPE_RRSIG,
@@ -194,6 +200,7 @@ ldns_dnssec_build_data_chain_other(ldns_resolver *res,
 		new_chain->signatures = signatures2;
 	}
 	ldns_pkt_free(my_pkt);
+	}
 }
 
 ldns_dnssec_data_chain *
@@ -223,6 +230,9 @@ ldns_dnssec_build_data_chain_nokeyname(ldns_resolver *res,
 	              LDNS_RR_TYPE_DS,
 	              LDNS_RR_CLASS_IN,
 	              qflags);
+	if (!my_pkt) {
+		return new_chain;
+	}
 
 	if (ldns_pkt_ancount(my_pkt) > 0) {
 		/* add error, no sigs but DS in parent */
@@ -352,8 +362,10 @@ ldns_dnssec_build_data_chain(ldns_resolver *res,
 			signatures = ldns_dnssec_pkt_get_rrsigs_for_type(pkt, type);
 		} else {
 			my_pkt = ldns_resolver_query(res, name, type, c, qflags);
+			if (my_pkt) {
 			signatures = ldns_dnssec_pkt_get_rrsigs_for_type(pkt, type);
 			ldns_pkt_free(my_pkt);
+			}
 		}
 	} else {
 		if (pkt) {
@@ -364,11 +376,13 @@ ldns_dnssec_build_data_chain(ldns_resolver *res,
 		}
 		if (!signatures) {
 			my_pkt = ldns_resolver_query(res, name, type, c, qflags);
+			if (my_pkt) {
 			signatures =
 				ldns_dnssec_pkt_get_rrsigs_for_name_and_type(my_pkt,
 													name,
 													type);
 			ldns_pkt_free(my_pkt);
+			}
 		}
 	}
 
@@ -415,6 +429,7 @@ ldns_dnssec_trust_tree_new()
 {
 	ldns_dnssec_trust_tree *new_tree = LDNS_XMALLOC(ldns_dnssec_trust_tree,
 										   1);
+        if(!new_tree) return NULL;
 	new_tree->rr = NULL;
 	new_tree->rrset = NULL;
 	new_tree->parent_count = 0;
@@ -481,6 +496,8 @@ ldns_dnssec_trust_tree_print_sm(FILE *out,
 	if (!sibmap) {
 		treedepth = ldns_dnssec_trust_tree_depth(tree);
 		sibmap = malloc(treedepth);
+                if(!sibmap)
+                        return; /* mem err */
 		memset(sibmap, 0, treedepth);
 		mapset = true;
 	}
@@ -638,6 +655,8 @@ ldns_dnssec_derive_trust_tree(ldns_dnssec_data_chain *data_chain, ldns_rr *rr)
 	size_t i, j;
 
 	ldns_dnssec_trust_tree *new_tree = ldns_dnssec_trust_tree_new();
+        if(!new_tree)
+                return NULL;
 	
 	if (data_chain && data_chain->rrset) {
 		cur_rrset = data_chain->rrset;
@@ -1116,12 +1135,9 @@ ldns_validate_domain_dnskey(const ldns_resolver * res,
 	ldns_rr_list * trusted_keys = NULL;
 
 	/* Fetch keys for the domain */
-	if ((keypkt = ldns_resolver_query(res,
-							    domain,
-							    LDNS_RR_TYPE_DNSKEY,
-							    LDNS_RR_CLASS_IN,
-							    LDNS_RD))) {
-
+	keypkt = ldns_resolver_query(res, domain,
+		LDNS_RR_TYPE_DNSKEY, LDNS_RR_CLASS_IN, LDNS_RD);
+	if (keypkt) {
 		domain_keys = ldns_pkt_rr_list_by_type(keypkt,
 									    LDNS_RR_TYPE_DNSKEY,
 									    LDNS_SECTION_ANSWER);
@@ -1205,12 +1221,9 @@ ldns_validate_domain_ds(const ldns_resolver *res,
 	ldns_rr_list * trusted_keys = NULL;
 
 	/* Fetch DS for the domain */
-	if ((dspkt = ldns_resolver_query(res,
-							   domain,
-							   LDNS_RR_TYPE_DS,
-							   LDNS_RR_CLASS_IN,
-							   LDNS_RD))) {
-
+	dspkt = ldns_resolver_query(res, domain,
+		LDNS_RR_TYPE_DS, LDNS_RR_CLASS_IN, LDNS_RD);
+	if (dspkt) {
 		rrset = ldns_pkt_rr_list_by_type(dspkt,
 								   LDNS_RR_TYPE_DS,
 								   LDNS_SECTION_ANSWER);
@@ -1411,6 +1424,10 @@ ldns_dnssec_verify_denial_nsec3(ldns_rr *rr,
 						   ldns_rr_owner(rr),
 						   ldns_rr_get_type(rr),
 						   nsecs);
+                if(!closest_encloser) {
+                        result = LDNS_STATUS_NSEC3_ERR;
+                        goto done;
+                }
 
 		wildcard = ldns_dname_new_frm_str("*");
 		(void) ldns_dname_cat(wildcard, closest_encloser);
@@ -1519,7 +1536,7 @@ ldns_gost2pkey_raw(unsigned char* key, size_t keylen)
 	memmove(encoded+37, key, 64);
 	pp = (unsigned char*)&encoded[0];
 
-	return d2i_PUBKEY(NULL, &pp, sizeof(encoded));
+	return d2i_PUBKEY(NULL, &pp, (int)sizeof(encoded));
 }
 
 static ldns_status
@@ -1545,6 +1562,63 @@ ldns_verify_rrsig_gost_raw(unsigned char* sig, size_t siglen,
 }
 #endif
 
+#ifdef USE_ECDSA
+EVP_PKEY*
+ldns_ecdsa2pkey_raw(unsigned char* key, size_t keylen, uint8_t algo)
+{
+	unsigned char buf[256+2]; /* sufficient for 2*384/8+1 */
+        const unsigned char* pp = buf;
+        EVP_PKEY *evp_key;
+        EC_KEY *ec;
+	/* check length, which uncompressed must be 2 bignums */
+        if(algo == LDNS_ECDSAP256SHA256) {
+		if(keylen != 2*256/8) return NULL;
+                ec = EC_KEY_new_by_curve_name(NID_X9_62_prime256v1);
+        } else if(algo == LDNS_ECDSAP384SHA384) {
+		if(keylen != 2*384/8) return NULL;
+                ec = EC_KEY_new_by_curve_name(NID_secp384r1);
+        } else    ec = NULL;
+        if(!ec) return NULL;
+	if(keylen+1 > sizeof(buf))
+		return NULL; /* sanity check */
+	/* prepend the 0x02 (from docs) (or actually 0x04 from implementation
+	 * of openssl) for uncompressed data */
+	buf[0] = POINT_CONVERSION_UNCOMPRESSED;
+	memmove(buf+1, key, keylen);
+        if(!o2i_ECPublicKey(&ec, &pp, (int)keylen+1)) {
+                EC_KEY_free(ec);
+                return NULL;
+        }
+        evp_key = EVP_PKEY_new();
+        if(!evp_key) {
+                EC_KEY_free(ec);
+                return NULL;
+        }
+        EVP_PKEY_assign_EC_KEY(evp_key, ec);
+        return evp_key;
+}
+
+static ldns_status
+ldns_verify_rrsig_ecdsa_raw(unsigned char* sig, size_t siglen, 
+	ldns_buffer* rrset, unsigned char* key, size_t keylen, uint8_t algo)
+{
+        EVP_PKEY *evp_key;
+        ldns_status result;
+        const EVP_MD *d;
+
+        evp_key = ldns_ecdsa2pkey_raw(key, keylen, algo);
+        if(!evp_key) {
+		/* could not convert key */
+		return LDNS_STATUS_CRYPTO_BOGUS;
+        }
+        if(algo == LDNS_ECDSAP256SHA256)
+                d = EVP_sha256();
+        else    d = EVP_sha384(); /* LDNS_ECDSAP384SHA384 */
+	result = ldns_verify_rrsig_evp_raw(sig, siglen, rrset, evp_key, d);
+	EVP_PKEY_free(evp_key);
+	return result;
+}
+#endif
 
 ldns_status
 ldns_verify_rrsig_buffers(ldns_buffer *rawsig_buf, ldns_buffer *verify_buf, 
@@ -1602,6 +1676,13 @@ ldns_verify_rrsig_buffers_raw(unsigned char* sig, size_t siglen,
 		return ldns_verify_rrsig_gost_raw(sig, siglen, verify_buf,
 			key, keylen);
 		break;
+#endif
+#ifdef USE_ECDSA
+        case LDNS_ECDSAP256SHA256:
+        case LDNS_ECDSAP384SHA384:
+		return ldns_verify_rrsig_ecdsa_raw(sig, siglen, verify_buf,
+			key, keylen, algo);
+		break;
 #endif
 	case LDNS_RSAMD5:
 		return ldns_verify_rrsig_rsamd5_raw(sig,
@@ -1708,6 +1789,17 @@ ldns_rrsig2rawsig_buffer(ldns_buffer* rawsig_buf, ldns_rr* rrsig)
 			return LDNS_STATUS_MEM_ERR;
 		}
 		break;
+#ifdef USE_ECDSA
+        case LDNS_ECDSAP256SHA256:
+        case LDNS_ECDSAP384SHA384:
+                /* EVP produces an ASN prefix on the signature, which is
+                 * not used in the DNS */
+		if (ldns_convert_ecdsa_rrsig_rdf2asn1(rawsig_buf, 
+			ldns_rr_rdf(rrsig, 8)) != LDNS_STATUS_OK) {
+			return LDNS_STATUS_MEM_ERR;
+                }
+                break;
+#endif
 	case LDNS_DH:
 	case LDNS_ECC:
 	case LDNS_INDIRECT:
